#include <floatType.h>
/****************************************************************************

COPYRIGHT NOTICE:

  The source code in this directory is provided free of
  charge to anyone who wants it.  It is in the public domain
  and therefore may be used by anybody for any purpose.  It
  is provided "AS IS" with no warranty of any kind
  whatsoever.
  
  Copyright: Bruno Ribeiro, UMass Amherst, 2006

****************************************************************************/


#include <stdio.h>
#include <math.h>
#include "wnlib.h"
#include "wnvect.h"

#include "wnconj.h"


local FLOAT s;
local int count;

#define __MAX__(a,b) (((a)>(b))?(a):(b))
#define __MIN__(a,b) (((a)>(b))?(b):(a))
/*#define ABS(a) (((a) > 0)?(a):(-a)) */
/*#define ABS(a) (a) */

/* Avoids negative solutions...*/
#define ABS(a) (((a) > 0)?(a):(-a*1e-06)) 

#define __PROB__ 0.005
#define __WMAX__ 50
// Same size as WMAX (sample size)
#define __SSIZE__ __WMAX__

#define LONGFL long double

#define INT unsigned int
#define READINT "%u"


// GLOBAL VAR
    FLOAT *gSYN;
    FLOAT *theta;
    FLOAT *b;
    FLOAT p = __PROB__;


//Returns the value ln[(xx)] for xx > 0.
FLOAT gammln(FLOAT xx)
{
//    Internal arithmetic will be done in FLOAT precision, a nicety that you can omit if five-figure accuracy is good enough.
    LONGFL x,y,tmp,ser;
    static LONGFL cof[6]={76.18009172947146,-86.50532032941677,
         24.01409824083091,-1.231739572450155,
         0.1208650973866179e-2,-0.5395239384953e-5};
    int j;
    
    y=x=xx;
    tmp=x+5.5;
    tmp -= (x+0.5)*logl(tmp);
    ser=1.000000000190015;
    for (j=0;j<=5;j++) ser += cof[j]/++y;
    return -tmp+logl(2.5066282746310005*ser/x);
}


FLOAT factln(INT n)
//Returns ln(n!).
{
//             A static array is automatically initialized to zero.
    static FLOAT a[101];
    if (n < 0) return -1e90;
    if (n <= 1) return 0.0;
//            In range of table.
    if (n <= 100) return a[n] ? a[n] : (a[n]=gammln(n+1.0));
//            Out of range of table.
    else return gammln(n+1.0);
}


FLOAT bico(INT n, INT k)
//Returns the binomial coefficient (n,k) as a floating-point number.
{
    return floor(0.5+EXPF(factln(n)-factln(k)-factln(n-k)));
//    The floor function cleans up roundoff error for smaller values of n and k.
}

FLOAT Logbico(INT n, INT k)
//Returns the binomial coefficient (n,k) as a floating-point number.
{
    return (factln(n)-factln(k)-factln(n-k));
//    The floor function cleans up roundoff error for smaller values of n and k.
}

FLOAT d(int j, FLOAT *x)
{
    int i;
    FLOAT dval = 0.0;
    
    for (i=j; i <= __WMAX__; i++)
    {
	    dval += b[(j-1)*__WMAX__ + i-1]*x[i-1];

    }
    return dval;
}



/* Rosenbrocks classic parabolic valley ("banana") function */
FLOAT L(FLOAT *thet)
{
    int j;
    FLOAT sum = 0.0;
  
    for (j=1; j <= __WMAX__; j++)
    {
       sum += gSYN[j-1] * LOGF(d(j,thet));
    }
    return (sum);
}

FLOAT diffL(int i, FLOAT *thet)
{
    int j;
    FLOAT sum = 0.0;

    for (j=1; j <= i; j++)
    {
       sum += b[(j-1)*__WMAX__ + i-1]*gSYN[j-1] / d(j,thet);
    }
    return (sum);
}


local FLOAT function(FLOAT v[])
{
  FLOAT ret;
  FLOAT sum=0.0;
  int i;
  
  for (i=0; i <__WMAX__; i++)
      sum += ABS(v[i]);
  for (i=0; i <__WMAX__; i++)
      v[i] = ABS(v[i])/sum;
  ret = L(v);
//  wn_print_vect(v,__WMAX__);
      
  return(-ret);
}


local void gradient(FLOAT grad[],FLOAT v[])
{
  int i;
  FLOAT sum=0.0;
  
  for (i=0; i <__WMAX__; i++)
      sum += ABS(v[i]);
  for (i=0; i <__WMAX__; i++)
      v[i] = ABS(v[i])/sum;
  if ((count % 100) == 0) printf("gradient. count=%d\n",count);
  ++count;

/*  wn_print_vect(v,__WMAX__);*/
  for (i=1; i<= __WMAX__; i++)
  {
  	grad[i-1] = diffL(i,v)-1.0;
  }
//  wn_print_vect(grad,__WMAX__);
/*  wn_print_vect(grad,__WMAX__);*/
}

local void computeB(FLOAT p)
{
   int i,j;
   
   b = malloc(sizeof(FLOAT)*__SSIZE__*__WMAX__);
   memset(b,0,sizeof(FLOAT)*__SSIZE__*__WMAX__);
   for (i = 1; i <= __WMAX__; i++)
   {
      for (j = 2; j <= i; j++)
      {
      // With SYN pkts
         b[(j-1)*__WMAX__ + i-1] = p*EXPF((i-j)*LOGF(1.0-p));
      }
      j = 1;
      b[(j-1)*__WMAX__ + i-1] = EXPF((i-j)*LOGF(1.0-p));
   }
      
}

local void readgSYN()
{
   int j;
   FLOAT gamma;
   FILE *f;
   
   f = fopen("gSYN.plot","r");
   if (f == NULL)
   {
       fprintf(stderr,"File gSYN.plot not found.");
       exit(1);
   }
   
   j = 0;
   gamma = 0;
   while (!feof(f))
   {
        if ((fscanf(f,PRINTF_FLT_e_,&gSYN[j]) > 0) && (j < __WMAX__))
        {
          gamma += (FLOAT)gSYN[j];
          j++;
        }
   }
   for (j = 1; j <= __WMAX__; j++)
   {
     gSYN[j-1] = gSYN[j-1]/gamma;
   }
   
   fclose(f);
}

local void removeB()
{
   free(b);
}

void file_print_vect(FILE *fout, FLOAT vect[],int len)
{
  int i;

  for(i=0;i<len;i++)
  {
    fprintf(fout, PRINTF_FLT_g_"\n",vect[i]);
  }
}


int main(void)
{
  FLOAT *vect;
  int code;
  int h;
  FLOAT val_min;
  FILE *fout;

  fout = fopen("inferredDist.plot","w");
  if (fout == NULL)
  {
       fprintf(stderr,"File inferredDist.plot could not be created.");
       exit(1);
  }
   
   
  wn_gpmake("no_free"); 

  wn_make_vect(&vect,__WMAX__);
  wn_make_vect(&gSYN,__WMAX__);
  wn_make_vect(&theta,__WMAX__);

  readgSYN();
  computeB(__PROB__);

  count = 0;

  for (h=0; h < __WMAX__; h++)
  {
     vect[h] = 1.0/__WMAX__;
  }

  s = 100.0;
  printf("s = "PRINTF_FLT_f_"\n",s);

  wn_conj_gradient_method(&code,&val_min,
		          vect,__WMAX__,(function),(gradient),50000);

  printf("solution %d\n",sizeof(FLOAT));
  file_print_vect(fout,vect,__WMAX__);
  printf("final result: code = %d   ",code);
  printf("count = %d\n",count);

  fclose(fout);
  
  removeB();
  wn_gpfree();
  
  return 0;
}
