/****************************************************************************

COPYRIGHT NOTICE:

  The source code in this directory is provided free of
  charge to anyone who wants it.  It is in the public domain
  and therefore may be used by anybody for any purpose.  It
  is provided "AS IS" with no warranty of any kind
  whatsoever.  For further details see the README files in
  the wnlib parent directory.

****************************************************************************/

/****************************************************************************

  "cdn" stands for conjugate-direction noisy

****************************************************************************/
#ifndef wncnjdn2H
#define wncnjdn2H


#include "wnlib.h"



#define WN_CDN_NOT_STARTED                -1



typedef struct wn_cdn_context_type_struct *wn_cdn_context_type;
typedef struct wn_cdn_srchdir_type_struct *wn_cdn_srchdir_type;

struct wn_cdn_context_type_struct
{
  int num_vars;

  FLOAT (*pfunction)(FLOAT vect[],int sequence_number);
  void (*pline_eval)(wn_cdn_context_type c,
                     FLOAT f[],FLOAT x[],int n,
                     FLOAT *search_direction);

  wn_cdn_srchdir_type *coord_search_direction_array,
                        *search_direction_array;
  int num_search_directions,max_num_search_directions;

  FLOAT *xa,*fa;
  int num_line_samples,min_num_line_samples,
      num_line_samples_mem;

  FLOAT *current_vect,*func_call_vect,
         *old_vect,*coord_direction;

  FLOAT ob;
  int code;

  int num_func_calls,max_num_func_calls;

  bool force_optimize_stop_flag /* = FALSE */;

  wn_memgp current_group;
};

struct wn_cdn_srchdir_type_struct
{
  bool is_coord_direction;
  int coord;
  FLOAT *dir_vect;

  FLOAT x_min,x_max;

  FLOAT x0;  /* optimimum x from last iteration */
  FLOAT curvature;  /* curvature from last iteration */
  FLOAT x_width,   /* x width from last iteration */
         max_x_width;
  
  FLOAT ratio_df1_noise;
};

EXTERN void wn_compute_func_call_vect
(
  wn_cdn_context_type c,
  FLOAT *search_direction,
  FLOAT x
);

EXTERN void wn_cdn_make_context
(
  wn_cdn_context_type *pc,
  FLOAT (*pfunction)(FLOAT vect[],int sequence_number),
  int num_vars
);

EXTERN void wn_cdn_free_context
(
  wn_cdn_context_type c
);

EXTERN void wn_cdn_set_solution
(
  wn_cdn_context_type c,
  FLOAT vect[]
);

EXTERN void wn_cdn_get_solution
(
  wn_cdn_context_type c,
  int *pcode,
  FLOAT *pob,
  FLOAT vect[]
);

EXTERN void wn_cdn_set_line_eval
(
  wn_cdn_context_type c,
  void (*pline_eval)(wn_cdn_context_type c,
                     FLOAT f[],FLOAT x[],int n,
                     FLOAT *search_direction)
);

EXTERN void wn_cdn_set_coord_x_min
(
  wn_cdn_context_type c,
  int coord,
  FLOAT x_min
);

EXTERN void wn_cdn_set_coord_x_max
(
  wn_cdn_context_type c,
  int coord,
  FLOAT x_max
);

EXTERN void wn_cdn_set_coord_x0
(
  wn_cdn_context_type c,
  int coord,
  FLOAT x0
);


EXTERN void wn_cdn_set_coord_x_width
(
  wn_cdn_context_type c,
  int coord,
  FLOAT x_width
);


EXTERN void wn_cdn_set_coord_max_x_width
(
  wn_cdn_context_type c,
  int coord,
  FLOAT max_x_width
);


EXTERN void wn_cdn_force_stop(wn_cdn_context_type c);


EXTERN void wn_cdn_optimize
(
  wn_cdn_context_type c,
  int num_func_calls
);



#endif

