/****************************************************************************

COPYRIGHT NOTICE:

  The source code in this directory is provided free of
  charge to anyone who wants it.  It is in the public domain
  and therefore may be used by anybody for any purpose.  It
  is provided "AS IS" with no warranty of any kind
  whatsoever.  For further details see the README files in
  the wnlib parent directory.

AUTHOR:

  Will Naylor

****************************************************************************/
#ifndef wnconjH
#define wnconjH


#include "wnlib.h"
#include "wncons.h"
#include "wnparvect.h"


#include "wnconjd.h"

typedef enum {WN_CONJ_ILLEGAL = 0,
	      WN_CONJ_DISTANCE, 
	      WN_CONJ_CURVATURE, 
	      WN_CONJ_MUL_FACTOR,
              WN_CONJ_NUM_ENTRIES} wn_jump_len_type;

EXTERN FLOAT wn_penalty(FLOAT x);
EXTERN FLOAT wn_dpenalty(FLOAT x);
EXTERN FLOAT wn_d2penalty(FLOAT x);

EXTERN FLOAT wn_barrier(FLOAT x);
EXTERN FLOAT wn_dbarrier(FLOAT x);

EXTERN FLOAT wn_barrier2(FLOAT x);
EXTERN FLOAT wn_dbarrier2(FLOAT x);

EXTERN FLOAT wn_clamp(FLOAT x);
EXTERN FLOAT wn_dclamp(FLOAT x);

EXTERN void wn_conj_gradient_method
(
  int *pcode,
  FLOAT *pval_min,
  FLOAT vect[],
  int len,
  FLOAT (*pfunction)(FLOAT vect[]),
  void (*pgradient)(FLOAT grad[],FLOAT vect[]),
  int max_iterations
);
EXTERN void wn_conj_gradient_diff_method
(
  int *pcode,
  FLOAT *pval_min,
  FLOAT vect[],
  FLOAT delta_vect[],
  int len,
  FLOAT (*pfunction)(FLOAT vect[]),
  int max_iterations
);
EXTERN void wn_conj_funcgrad_method
(
  int *pcode,
  FLOAT *pval_min,
  FLOAT vect[],
  int len,
  FLOAT (*pfuncgrad)(FLOAT grad[],FLOAT vect[]),
  bool (*pterminate)(FLOAT vect[],FLOAT best_f,int iteration),
  FLOAT first_jump_len,
  wn_jump_len_type jump_len_flag,
  int no_reset_run_len,
  FLOAT reset_expand_factor,
  int max_iterations
);
EXTERN void wn_conj_funcgrad_method_parallel
(
  int *pcode,
  FLOAT *pval_min,
  FLOAT vect[],
  int num_vars,
  FLOAT (*pfuncgrad)(FLOAT grad[],FLOAT vect[]),
  bool (*pterminate)(FLOAT vect[],FLOAT best_f,int iteration),
  FLOAT first_step_len,
  wn_jump_len_type jump_len_flag,
  int no_reset_run_len,
  FLOAT reset_expand_factor,
  int max_iterations,
  wn_parallel_callback par_cb,
  int num_threads
);
EXTERN void wn_lbfgs_funcgrad_method
(
  int *pcode,
  FLOAT *pval_min,
  FLOAT vect[],
  int num_vars,
  FLOAT (*pfuncgrad)(FLOAT grad[],FLOAT vect[]),
  bool (*pterminate)(FLOAT vect[],FLOAT best_f),
  FLOAT first_jump_len,
  int max_depth, 
  int max_iterations
);

EXTERN void wn_force_conj_gradient_stop(void);

EXTERN void wn_numerical_gradient
(
  FLOAT grad[],
  FLOAT vect[],
  FLOAT delta_vect[],
  int len,
  FLOAT (*pfunction)(FLOAT vect[])
);


EXTERN void wn_fit_parabola_3p
(
  int *pcode,
  FLOAT *pa,FLOAT *px0,FLOAT *pb,
  FLOAT x1,FLOAT y1,
  FLOAT x2,FLOAT y2,
  FLOAT x3,FLOAT y3
);
EXTERN void wn_fit_parabola_2pd
(
  int *pcode,
  FLOAT *pa,FLOAT *px0,FLOAT *pb,
  FLOAT x1,FLOAT y1,FLOAT dy1,
  FLOAT x2,FLOAT y2
);
EXTERN void wn_fit_parabola_2d
(
  int *pcode,
  FLOAT *pa,FLOAT *px0,
  FLOAT x1,FLOAT dy1,
  FLOAT x2,FLOAT dy2
);
EXTERN void wn_fit_traditional_parabola_3p
(
  int *pcode,
  FLOAT *pa,FLOAT *pb, FLOAT *pc,
  FLOAT x1,FLOAT y1,
  FLOAT x2,FLOAT y2,
  FLOAT x3,FLOAT y3
);
EXTERN void wn_fit_traditional_parabola_2pa
(
  int *pcode,
  FLOAT *pb, FLOAT *pc,
  FLOAT a,
  FLOAT x1,FLOAT y1,
  FLOAT x2,FLOAT y2
);
EXTERN void wn_convert_parabola_traditional_to_centered
(
  int *pcode,
  FLOAT *px0,FLOAT *pb,
  FLOAT a,
  FLOAT b,
  FLOAT c
);
EXTERN void wn_fit_cubic_2p2d
(
  int *pcode,
  FLOAT coef[4],
  FLOAT x1,FLOAT y1,FLOAT dy1,
  FLOAT x2,FLOAT y2,FLOAT dy2
);

EXTERN FLOAT wn_clip_f(FLOAT);


EXTERN void wn_minimize_1d
(
  int *pcode,
  FLOAT *pval_min,
  FLOAT *px_min,
  FLOAT (*pfunction)(FLOAT x),
  int max_iterations
);
EXTERN void wn_minimize_1d_raw
(
  int *pcode,
  FLOAT *pf0,
  FLOAT *pf1,
  FLOAT *pf2,
  FLOAT *px0,
  FLOAT *px1,
  FLOAT *px2,
  FLOAT f_goal,
  FLOAT (*pfunction)(FLOAT x),
  int max_improve_iterations,
  int max_total_iterations
);


#endif

